from PIL import Image, ImageOps, ImageChops
from tqdm import tqdm
from diffusers import StableDiffusionInpaintPipeline
from diffusers.utils import load_image, make_image_grid
import pandas as pd
import random, math, torch, lpips, os
import torchvision.transforms as transforms



class SDInpainting:

    def __init__(self, image_png, model: str, prompt: str, pipe, seed: int, infsteps: int):
        self.orig_image, self.mask_orig_image, self.nmask_orig_image = self.adjust_size(image_png)
        self.generator = torch.Generator(device="cuda").manual_seed(seed)
        self.random = random.Random(seed)
        self.result_image = None
        self.prompt = prompt
        self.infsteps = infsteps
        self.pipe = pipe
        if not pipe:
            self.pipe = StableDiffusionInpaintPipeline.from_pretrained(
                    model,
                    torch_dtype=torch.float16,
                    token=access_token,
                    cache_dir="/home/jovyan/cached_models"
            )
            self.pipe.to("cuda")    
        self.model_conf = {"num_inference_steps": self.infsteps}


    def generate(self, image=None, mask_image=None, **kwargs):
        image = ImageChops.lighter(image, mask_image)
        return self.pipe(image=image, mask_image=mask_image, **kwargs)


    def apply_mask(self, pot=0, random_selection=True, last_selected_block=0):
        random_mask = Image.new("RGB", self.mask_orig_image.size, (0, 0, 0))
        width, height = random_mask.size
        step = 2 ** pot  # Cálculo del paso basado en la potencia
    
        # Determine the total number of blocks in width and height
        num_blocks_x = (width + step - 1) // step
        num_blocks_y = (height + step - 1) // step
        total_blocks = num_blocks_x * num_blocks_y
        #print(total_blocks)
    
        if random_selection:
            # Select a random block
            selected_block_index = self.random.randint(0, total_blocks - 1)
        else:
            # Select the block following the last selected block
            selected_block_index = (int(last_selected_block) + 1) % total_blocks
    
        # Convert the block index to block coordinates
        selected_block_x = selected_block_index % num_blocks_x
        selected_block_y = selected_block_index // num_blocks_x

        random_mask = ImageChops.lighter(self.nmask_orig_image, random_mask)
        # Color the selected block with white
        for dx in range(step):
            for dy in range(step):
                if (selected_block_x * step + dx < width) and (selected_block_y * step + dy < height):
                    random_mask.putpixel((selected_block_x * step + dx, selected_block_y * step + dy), (255, 255, 255))
        random_mask = ImageChops.darker(self.mask_orig_image, random_mask)
        return random_mask, selected_block_index


    def apply_regeneration(self, image1=None, image2=None, mask=None):
        if not image1: # black zone
            image1 = self.orig_image.copy()
        if not image2: # white zone
            image2 = self.result_image.copy()
        if not mask:
            mask = self.nmask_orig_image.copy()
        fn = lambda x : 255 if x > 127 else 0
        mask = mask.convert('L').point(fn, mode='1')
        result = Image.composite(image2, image1, mask)
        return result
    
    
    def adjust_size(self, image, target_size=(512,512)):
        mask_image = Image.new("RGB", image.size, (255, 255, 255))
        nmask_image = Image.new("RGB", image.size, (0, 0, 0))
        new_image = ImageOps.pad(image, target_size, Image.LANCZOS, (255, 255, 255), (0.5,0.5))
        mask_image = ImageOps.pad(mask_image, target_size, Image.LANCZOS, (100, 100, 100), (0.5,0.5))
        nmask_image = ImageOps.pad(nmask_image, target_size, Image.LANCZOS, (100, 100, 100), (0.5,0.5))
        return new_image, mask_image, nmask_image

    
    def make_grid(self, images: list):
        n_images = len(images)
        rows = int(math.sqrt(n_images)+0.5)
        cols = int(math.ceil(n_images/rows))
        n_fill_images = rows*cols-n_images
        for i in range(n_fill_images):
            images.append(Image.new("RGB", self.orig_image.size, (255, 255, 255)))
        return make_image_grid(images, rows, cols)


    def lpips_distance(self, img1, img2):
        tensor_img1 = transforms.ToTensor()(img1)
        tensor_img2 = transforms.ToTensor()(img2)
        loss_fn_alex = lpips.LPIPS(net='alex', verbose=False) # best forward scores
        loss_fn_vgg = lpips.LPIPS(net='vgg', verbose=False) # closer to "traditional" perceptual loss 
        loss_fn_squeeze = lpips.LPIPS(net='squeeze', verbose=False)
        return loss_fn_alex(tensor_img1, tensor_img2).item(), loss_fn_vgg(tensor_img1, tensor_img2).item(), loss_fn_squeeze(tensor_img1, tensor_img2).item()


    def save_image(self, path, image):
        directory = os.path.dirname(path)
        if not os.path.exists(directory):
            os.makedirs(directory)
        image.save(path)
        

    def experiment1(self, image_id, resolutions: list, seed: int, save_path: str):
        num_iters = 16
        step = 2
        image_id_str = f"{image_id:02d}"
        df_distances = pd.DataFrame(columns=['image_id', 'image_id_str', 'seed', 'mask', 'r', 'changed_percentage', 'changed', 'alex_lpips_orig', 'alex_lpips_prev', 'vgg_lpips_orig', 'vgg_lpips_prev', 'squeeze_lpips_orig', 'squeeze_lpips_prev'])
        pbar_resolutions = tqdm(range(len(resolutions)), desc=f'image_{image_id_str}: ', colour='green', leave=False)
        pbar_resolutions.n = 0
        pbar_resolutions.refresh()
        for resolution in resolutions:
            mask_bits = 2**resolution
            mask_str = f'{mask_bits}x{mask_bits}'
            changed = 0.5
            changed_percentage = int(changed*100)
            image_id_str = f"{image_id:02d}_s{seed}_{mask_str}"
            self.random.seed(seed)
            generator = torch.Generator(device="cuda").manual_seed(seed)
            distance_0 = self.lpips_distance(self.orig_image, self.orig_image)
            df_distances.loc[len(df_distances)] = [image_id, image_id_str, seed, mask_str, resolution, 0, 0] + list(distance_0) + list(distance_0)
            self.result_image = self.orig_image.copy()
            prev_result = self.orig_image.copy()
            grid_images = [self.result_image]
            pbar_iters = tqdm(range(num_iters), desc=f'mask {mask_str}: ', colour='green', leave=False)
            pbar_iters.n = 0
            pbar_iters.refresh()
            for i in range(1, num_iters+1):
                mask,n = self.apply_mask(resolution)
                self.result_image = self.generate(prompt=self.prompt, image=self.result_image, mask_image=mask, generator=generator, **self.model_conf).images[0]
                self.result_image = self.apply_regeneration(mask=mask, image1=prev_result)
                if (int(i % step) == 0):
                    grid_images.append(self.result_image)
                    distance_orig = self.lpips_distance(self.orig_image, self.result_image)
                    distance_prev = self.lpips_distance(self.result_image, prev_result)
                    df_distances.loc[len(df_distances)] = [image_id, image_id_str, seed, mask_str, resolution, changed_percentage, changed, distance_orig[0], distance_prev[0], distance_orig[1], distance_prev[1], distance_orig[2], distance_prev[2]]
                    changed += 0.5
                prev_result = self.result_image.copy()
                pbar_iters.n += 1
                pbar_iters.refresh()
                
            grid = self.make_grid(grid_images)
            self.save_image(f'{save_path}/mask_{mask_str}/{image_id_str}.png', grid)
            num_iters *= 4
            step *= 4
            pbar_resolutions.n += 1
            pbar_resolutions.refresh()
        return df_distances


    def experiment2(self, image_id, resolutions: list, seeds: list, save_path: str):
        num_iters = 16
        step = 2
        image_id_str = f"{image_id:02d}"
        df_distances = pd.DataFrame(columns=['image_id', 'image_id_str', 'seed', 'mask', 'r', 'changed_percentage', 'changed', 'alex_lpips_orig', 'alex_lpips_prev', 'vgg_lpips_orig', 'vgg_lpips_prev', 'squeeze_lpips_orig', 'squeeze_lpips_prev'])
        pbar_seeds = tqdm(range(len(seeds)), desc=f'image_{image_id_str}: ', colour='green', leave=False)
        pbar_seeds.n = 0
        pbar_seeds.refresh()
        for seed in seeds:
            pbar_resolutions = tqdm(range(len(resolutions)), desc=f'seed_{seed}: ', colour='green', leave=False)
            pbar_resolutions.n = 0
            pbar_resolutions.refresh()
            for resolution in resolutions:
                mask_bits = 2**resolution
                mask_str = f'{mask_bits}x{mask_bits}'
                changed = 0.5
                changed_percentage = int(changed*100)
                image_id_str = f"{image_id:02d}_s{seed}_{mask_str}"
                self.random.seed(seed)
                generator = torch.Generator(device="cuda").manual_seed(seed)
                distance_0 = self.lpips_distance(self.orig_image, self.orig_image)
                df_distances.loc[len(df_distances)] = [image_id, image_id_str, seed, mask_str, resolution, 0, 0] + list(distance_0) + list(distance_0)
                self.result_image = self.orig_image.copy()
                prev_result = self.orig_image.copy()
                grid_images = [self.result_image]
                pbar_iters = tqdm(range(num_iters), desc=f'mask {mask_str}: ', colour='green', leave=False)
                pbar_iters.n = 0
                pbar_iters.refresh()
                for i in range(1, num_iters+1):
                    mask,n = self.apply_mask(resolution)
                    self.result_image = self.generate(prompt=self.prompt, image=self.result_image, mask_image=mask, generator=generator, **self.model_conf).images[0]
                    self.result_image = self.apply_regeneration(mask=mask, image1=prev_result)
                    if (int(i % step) == 0):
                        grid_images.append(self.result_image)
                        distance_orig = self.lpips_distance(self.orig_image, self.result_image)
                        distance_prev = self.lpips_distance(self.result_image, prev_result)
                        df_distances.loc[len(df_distances)] = [image_id, image_id_str, seed, mask_str, resolution, changed_percentage, changed, distance_orig[0], distance_prev[0], distance_orig[1], distance_prev[1], distance_orig[2], distance_prev[2]]
                        changed += 0.5
                    prev_result = self.result_image.copy()
                    pbar_iters.n += 1
                    pbar_iters.refresh()
    
                grid = self.make_grid(grid_images)
                self.save_image(f'{save_path}/mask_{mask_str}/{image_id_str}.png', grid)
                num_iters *= 4
                step *= 4
                pbar_resolutions.n += 1
                pbar_resolutions.refresh()
            pbar_seeds.n += 1
            pbar_seeds.refresh()
        return df_distances