from PIL import Image, ImageOps, ImageChops
from datasets import load_dataset, load_from_disk
from tqdm import tqdm
import random
import pandas as pd
import numpy as np
import matplotlib.pyplot as plt
import math, requests, argparse, os, gc, time, traceback, subprocess, zipfile


def compress_folder(path, zipname):
    with zipfile.ZipFile(zipname, 'w', zipfile.ZIP_DEFLATED) as zipfile:
        for root_folder, sub_folders, files in os.walk(path):
            for file in files:
                complete_path = os.path.join(root_folder, file)
                relative_file = os.path.relpath(complete_path, path)
                zipfile.write(complete_path, relative_file)


def prepare_dataset(args):
    cmd = f'python load_images_dataset.py -l 100 -s 2 -grows 10 -gcols 10 -oi {args.output}/original_imgs_grid.png -oim {args.output} --save-metadata'
    p = subprocess.Popen(cmd, stdout=subprocess.PIPE, shell=True)
    p.wait()
    time.sleep(5)
    return load_from_disk(args.input).select(range(args.first, args.first+args.length))



if __name__ == '__main__':

    parser = argparse.ArgumentParser()

    parser.add_argument(
        "-c",
		"--cuda",
		type=str,
        default="0",
		required=False,
		help='CUDA_VISIBLE_DEVICES',
	)

    parser.add_argument(
        "-i",
		"--input",
		type=str,
        default="wikiart_test",
		required=False,
		help="input images dataset path",
	)

    parser.add_argument(
        "-p",
		"--prompt",
		type=str,
        default="",
		required=False,
		help="prompt",
	)
    
    parser.add_argument(
        "-f",
		"--first",
		type=int,
        default=0,
		required=False,
		help="initial image index",
	)

    parser.add_argument(
        "-l",
		"--length",
		type=int,
        default=100,
		required=False,
		help="total images",
	)

    parser.add_argument(
        "-if",
		"--infsteps",
		type=int,
        default=50,
		required=False,
		help="inference steps",
	)

    parser.add_argument(
        "-s",
		"--seed",
		type=int,
        default=0,
		required=False,
		help="seed",
	)
    
    parser.add_argument(
        "-o",
		"--output",
		type=str,
        default="lpips_distances_100_images_1_seed",
		required=False,
		help="destination path",
	)

    args = parser.parse_args()
    
    os.environ['CUDA_VISIBLE_DEVICES'] = args.cuda
    os.environ['TF_CPP_MIN_LOG_LEVEL'] = '2'
    os.environ['TF_ENABLE_ONEDNN_OPTS'] = '0'

    import torch, lpips
    import torchvision.transforms as transforms
    from diffusers import StableDiffusionInpaintPipeline
    from diffusers.utils import load_image, make_image_grid
    from diffusers.utils.logging import disable_progress_bar
    from SDInpainting import SDInpainting

    dataset = prepare_dataset(args)
    model = "stabilityai/stable-diffusion-2-inpainting"
    random.seed(args.seed)
    df = pd.DataFrame(columns=['image_id', 'image_id_str', 'seed', 'r', 'mask', 'changed', 'alex_lpips_orig', 'alex_lpips_prev', 'vgg_lpips_orig', 'vgg_lpips_prev', 'squeeze_lpips_orig', 'squeeze_lpips_prev'])
    pipe = StableDiffusionInpaintPipeline.from_pretrained(
                model,
                torch_dtype=torch.float16,
                cache_dir="/home/jovyan/cached_models"
        )
    pipe.to("cuda")
    pipe.set_progress_bar_config(disable=True)
    
    try:
        pbar_images = tqdm(range(len(dataset)), desc=f'Experiment progress: ', colour='green')
        pbar_images.n = 0
        pbar_images.refresh()
        for i,img in enumerate(dataset['image']):
            obj = SDInpainting(img, model, args.prompt, pipe, args.seed, args.infsteps)
            df = pd.concat([df, obj.experiment1(args.first+i, [8,7,6], args.seed, f'{args.output}')])
            df.to_excel(f"{args.output}/lpips_distances.xlsx", index=False, engine="openpyxl")
            pbar_images.n += 1
            pbar_images.refresh()
        
    except (Exception, KeyboardInterrupt) as e:
        traceback.print_exc()
        if obj:
            print("\nSaving data and progress...")
            df.to_excel(f"{args.output}/lpips_distances.xlsx", index=False, engine="openpyxl")
            print("Saved!")
    finally:
        df.to_excel(f"{args.output}/lpips_distances.xlsx", index=False, engine="openpyxl")
        compress_folder("lpips_distances_100_images_1_seed", "lpips_distances_100_images_1_seed.zip")
        del obj
        gc.collect()
        torch.cuda.empty_cache()
    time.sleep(5)